# The MIT License (MIT)
#
# Copyright (c) 2018 James K Bowler, Data Centauri Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE..

from opencollect.reader import SocketsStreamReader
from opencollect.writer import DatabaseWriter


class OpenCollect(object):
    def __init__(self):
        self.reader = SocketsStreamReader()
        self.writer = DatabaseWriter

    def start(self):
        while True:
            data = self.reader.read_buffer()
            if data:
                if data.payload == "exit":
                    self.reader.close_session()
                    break
                write = self.writer(data)
                write.start()


# The MIT License (MIT)
#
# Copyright (c) 2018 James K Bowler, Data Centauri Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

#from openconnect.settings import HOST, USER, PASSWD
from threading import Thread
import pymysql
pymysql.install_as_MySQLdb()
import MySQLdb


class DatabaseWriter(Thread):
    """
    A class representing a threaded DatabaseWriter.
    It contains an id for tracking with the ability
    to handle different types of database sql statements

    Parameters
    ----------
    container  : NamedTuple
    
    """
    def __init__(self, container):
        super().__init__()
        self.id = container.identifer
        self._container = container
        #self._h = HOST
        #self._u = USER
        #self._p = PASSWD
        #self._con = self._db_connection()
        #self._cur = self._con.cursor()

    def _db_connection(self):
        return MySQLdb.connect(
                host=self._h,
                user=self._u,
                passwd=self._p
            )

    def _execute_many(self, stmt, payload):
        try:
            print(self.id, "stmt", "payload")
            #self._cur.executemany(stmt, payload)
            #self._con.commit()
        finally:
            #self._cur.close()
            #self._con.close()          
            pass

    def run(self):
        stmt = self._container.query
        payload = self._container.payload
        self._execute_many(stmt, payload)


# The MIT License (MIT)
#
# Copyright (c) 2018 James K Bowler, Data Centauri Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from datetime import datetime, timedelta
from random import randint

import socket 
import json
import struct


class ClientExample(object):
    def __init__(self):
        self.sock = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM
        )
        self.ip = "127.0.0.1"
        self.port = 60312
        self.dummy_payload = []
        self.dt = datetime.utcnow()
        insert = "REPLACE INTO {0}.{1} ".format('db_name', 'tb_name')
        stmt = """(date, open, high, low, close, volume
                  ) VALUES (%s, %s, %s, %s, %s, %s);"""
        self._sql = insert + stmt
        
    def dummy(self):
        dummy_payload = []
        rand = randint(1, 300)
        for i in range(rand):
            self.dt+=timedelta(minutes=1)
            data = [datetime.strftime(self.dt, '%Y-%m-%d %H:%M:%S.%f'), 1.3991 , 1.40706, 1.39288, 1.39702, 306833]
            dummy_payload.append(data)
        return dummy_payload
    
    def get_client_socket(self):
        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        client.connect((self.ip, self.port))
        return client
    
    def create_message(self, id, payload, sql):
        identifer = "{}_{}".format(self.__class__.__name__, str(id))
        msg = {"identifer": identifer, "payload": payload, "query": sql}
        message = json.dumps(msg).encode()
        message = struct.pack('>I', len(message)) + message
        return message

    def send_exit_msg(self):
        client = self.get_client_socket()
        message = self.create_message(0, "exit", "None")
        client.sendall(message)
        client.recv(4096)
        
    def data(self):
        for i in range(1, 1000):
            client = self.get_client_socket()
            message = self.create_message(i, self.dummy(), self._sql)
            client.sendall(message)
            client.recv(4096)
        self.send_exit_msg()
        
    def run(self):
        self.data()
        
        
ClientExample().run()


__author__ = "James K Bowler"
__copyright__ = "Copyright 2018"
__license__ = "MIT"
__version__ = "1.2.1a2"
__maintainer__ = "JamesKbowler"
__email__ = "james.bowler@datacentauri.com"
__status__ = "Development"
__url__ = "https://github.com/JamesKBowler/opencollect"


# The MIT License (MIT)
#
# Copyright (c) 2018 James K Bowler, Data Centauri Ltd
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from collections import namedtuple
from threading import Thread
import json
import time
import socket
import struct


class SocketsStreamReader(object):
    """
    A class representing a none blocking SocketsStreamReader.
    It contains an id for tracking with the ability
    to handle different types of streamed events.

    Parameters
    ----------
    port       : TCP port number
    host       : IP Address
    identifier : human-readable identifier
    
    Exposes
    -------
    read_buffer()
    
    """
    def __init__(
        self,
        port=60312,
        host="0.0.0.0",
        identifer="Terminator"
    ):
        self.id = identifer
        self._buffer = []
        self._stream = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._stream.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        
        self._stream.bind((host, port))
        self._stream.listen(5)
        
        def _stream_to_buffer(id, stream, buffer):
            """
            Stream reading.
            """
            def handle(id, sock, buffer):
                def recv_msg(sock):
                    """
                    Get the message
                    """
                    raw_msglen = recvall(sock, 4)
                    if not raw_msglen:
                        return None
                    msglen = struct.unpack('>I', raw_msglen)[0]
                    return recvall(sock, msglen)

                def recvall(sock, n):
                    """
                    Piece large data
                    """
                    data = b''
                    while len(data) < n:
                        packet = sock.recv(n - len(data))
                        if not packet:
                            return None
                        data += packet
                    return data
                
                def container(result):
                    """
                    Container for payload
                    """
                    sm = json.loads(result.decode())
                    if not isinstance(sm, dict):
                        raise AttributeError("Loaded json must be a dict")
                    keys = sm.keys()
                    if 'identifer' not in keys:
                        raise KeyError("no identifer found!")
                    if 'payload' not in keys:
                        raise KeyError("no payload found!")
                    if 'query' not in keys:
                        raise KeyError("no query found!")
                    if len(keys) > 3:
                        raise AttributeError("Unexpected key in dict")
                    DataTuple = namedtuple(sm['identifer'], ['identifer', 'payload', 'query'])
                    return DataTuple(**sm)

                result = recv_msg(sock)
                sock.send(b'ACK!')
                sock.close()
                if result:
                    buffer.append(container(result))
                else:
                    raise EOFError(
                        "Unexpected End of Stream {}".format(id)
                    )

            while True:
                try:
                    sock = stream.accept()[0]
                except OSError:
                    print("Shutting down SocketsStreamReader")
                    break
                client_handler = Thread(
                    target=handle,
                    args=(id, sock, buffer)
                )
                client_handler.start()

        self.t = Thread(
            target=_stream_to_buffer,
            args=(self.id, self._stream, self._buffer,)
        )
        self.t.daemon = True
        self.t.start()

    def close_session(self):
        """
        Close socket
        """
        self._stream.shutdown(socket.SHUT_WR)
        time.sleep(2)
        self._stream.close()
        
    def read_buffer(self, sleepy_time=0.00001):
        """
        Read items first item and removes from the
        the list. (FIFO)
        
        Returns
        -------
        namedtuple
        
        """
        try:
            return self._buffer.pop(0)
        except IndexError:
            return time.sleep(sleepy_time)
